/* ShawEyes.ino:  Copyright 2017 Jetty

   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <WebServer.h>

// RGB LED flashes white every 10 seconds for 0.5 seconds, indicating a successful ping
// RGB LED flashes red every 10 seconds for 0.5 seconds, indicating a failed ping
// RGB lED breathes cyan when connected to the cloud
// RGB LED breathes green when not connected to the cloud
// D7 LED is blue when modem is being provided power, and off when it's not.  It will switch off and on for 4 seconds during a power cycle
// Auto power cycling is limited to a 20 min interval to prevent excessive cycling
//
// IMPORTANT UPDATING:  This device boots disconnected from the cloud.  To connect to the cloud to update firmware,
// either put in Safe Mode, or connect to the device via a local web browser and switch Cloud On, then flash the firmware.

SYSTEM_MODE(SEMI_AUTOMATIC);

#define SIMULATE_INTERNET_FAILURE

#define NUM_PING_PACKETS        5
#define LED_PIN                 D7
#define INTERNET_POWER_PIN      D6
#define TZ_OFFSET              -7

static  IPAddress               gatewayIP(183, 69, XXX, XXX);	// CHANGE THE IP HERE TO YOUR GATEWAY IP
static  WebServer               webserver("", 80);
static  int                     numPacketsReceived;
static  int                     bootTime;
static  int                     lastPingTime;
static  String                  macAddress;
static  int                     logSize;

#ifdef SIMULATE_INTERNET_FAILURE
    static int internetFailure = 0;
#endif

const char *logEntryActionTypeNames[] = {   
                                            "None", 
                                            "Starting Monitoring",
                                            "Connection Lost",
                                            "Connection Recovered",
                                            "Disabling Connection Reporting Till Active",
                                            "Internet Auto-Restarted !",
                                            "Manual Internet Restart",
                                            NULL
                                        };

#pragma pack(1)

enum logEntryActionType
{
    LEAT_NONE = 0,
    LEAT_BOOT,
    LEAT_CONNECTION_LOST,
    LEAT_CONNECTION_RESUMED,
    LEAT_DISABLE_CONNECTION_REPORT_TILL_ACTIVE,
    LEAT_RESTART_INTERNET_AUTO,
    LEAT_RESTART_INTENERT_MANUAL
};

struct logEntry
{
    int     entryTime;
    byte    entryAction;
};

#define LOG_ENTRY_OFFSET        (sizeof(logSize))
#define LOG_ENTRY_SIZE          (sizeof(struct logEntry))

#define POWER_CYCLE_OFF_TIME_MS         4000
#define REPEATED_RESET_LOCKOUT_TIME     (20 * 60)   // 20 mins
#define MAX_FAILED_PING_COUNT           3           // How many failed pings before we reboot
#define STARTUP_LOCKOUT_TIME            (10 * 60)   // 10 mins - How long to ignore lost connections at startup.  This is
                                                    // because restarting the Photon will cause a loss of internet connection, we
                                                    // don't want to double up on resets on startup



void putLogSize(void)
{
    EEPROM.put(0, logSize);
}



void resetLog()
{
    logSize = 0;
    putLogSize();
}



int getLogSize(void)
{
    EEPROM.get(0, logSize);
    return logSize;
}



// Returns 1 if a log entry is returned, otherwise 0 when the end of the log has been found

int getNextLogEntry(int entryIndex, struct logEntry *logEntry)
{
    int addr;

    addr = LOG_ENTRY_OFFSET + (entryIndex * LOG_ENTRY_SIZE);

    if ( entryIndex >= logSize )   return 0;
    
    // If we've exceeded the size of eeprom, return
    if (( addr + (int)LOG_ENTRY_SIZE ) >= ((int)EEPROM.length() - 1)) return 0;
     
    EEPROM.get(addr, *logEntry);

    addr += LOG_ENTRY_SIZE;

    return 1;
}



// If log is full, this empties the log first

void addLogEntry(struct logEntry logEntry)
{
    int addr;
    
    getLogSize();
    
    addr = (logSize * LOG_ENTRY_SIZE) + LOG_ENTRY_OFFSET;
    
    // If we don't have enough room, clear log
    if (( addr + (int)LOG_ENTRY_SIZE ) >= ((int)EEPROM.length() - 1))
    {
        resetLog();
        addr = LOG_ENTRY_OFFSET;
    }
    
    EEPROM.put(addr, logEntry);
    
    logSize ++;
    putLogSize();
}



// Returns 1 if successfull, otherwise 0

int powerCycleInternet(int automatic)
{
    struct logEntry logEntry;
    static int lastSuccessfulReset = 0;
    
    logEntry.entryTime      = Time.now();
    
    if ( automatic && ( logEntry.entryTime < (lastSuccessfulReset + REPEATED_RESET_LOCKOUT_TIME)) )
        return 0;

    logEntry.entryAction    = ( automatic ) ? LEAT_RESTART_INTERNET_AUTO : LEAT_RESTART_INTENERT_MANUAL;
    addLogEntry(logEntry);

    digitalWrite(LED_PIN, LOW);
    digitalWrite(INTERNET_POWER_PIN, LOW);

    delay(POWER_CYCLE_OFF_TIME_MS);

    digitalWrite(LED_PIN, HIGH);
    digitalWrite(INTERNET_POWER_PIN, HIGH);
    
    lastSuccessfulReset = Time.now();
    
    return 1;
}



// index.html
void web_index(WebServer &server, WebServer::ConnectionType type, char *, bool)
{
    struct logEntry logEntry;
    
    server.httpSuccess();
    server.print("<!DOCTYPE html><html><head><title>Shaw Eyes</title><body><H1>Shaw Eyes</H1><table>");
    server.print("<tr><td>SSID:</td><td>" + String(WiFi.SSID()) + "</td></tr>");
    server.print("<tr><td>RSSI:</td><td>" + String(WiFi.RSSI()) + " dB (-1dB is strongest signal)</td></tr>");
    server.print("<tr><td>IP:</td><td>" + String(WiFi.localIP()) + "</td></tr>");
    server.print("<tr><td>Mac:</td><td>" + macAddress + "</td></tr>");
    server.print("<tr><td>Eprom Size(physical):</td><td>" + String(EEPROM.length()) + " bytes</td></tr>");
    server.print("<tr><td># Log Entries:</td><td>" + String(getLogSize()) + " of " + ((EEPROM.length() - (1 + LOG_ENTRY_OFFSET)) / LOG_ENTRY_SIZE) + "</td></tr>");
    server.print("<tr><td>Startup time:</td><td>" + Time.format(bootTime, TIME_FORMAT_DEFAULT) + "</td></tr>");
    server.print("<tr><td>Last ping:</td><td>" + Time.format(lastPingTime, TIME_FORMAT_DEFAULT) + " returned " + String(numPacketsReceived) + " packets</td></tr>");
    server.print("</table><H2>Restart Log:</H2>");
    
    server.print("<table>");
    getLogSize();
    for ( int i = logSize-1; i >= 0 && getNextLogEntry(i, &logEntry); i-- )
    {
        server.print("<tr><td>" + Time.format(logEntry.entryTime, TIME_FORMAT_DEFAULT) + "</td><td>" + String(logEntryActionTypeNames[logEntry.entryAction]) + "</td></tr>");
    }
    server.print("</table>");
    
    server.print("<p /><table><tr>");
    server.print("<td><input type=\"button\" onclick=\"location.href='restartModem';\" value=\"Power Cycle Modem\" /></td>");
    server.print("<td><input type=\"button\" onclick=\"location.href='resetLog';\" value=\"Clear Log\" /></td>");
    server.print("<td><input type=\"button\" onclick=\"location.href='reset';\" value=\"Reset Photon\" /></td>");

#ifdef SIMULATE_INTERNET_FAILURE
    server.print("<td><input type=\"button\" onclick=\"location.href='toggleInternet';\" value=\"Switch Internet " + String(( internetFailure ) ?"On" : "Off") + "\" /></td>");
#endif

    server.print("<td><input type=\"button\" onclick=\"location.href='toggleCloud';\" value=\"Switch Cloud " + String(( Particle.connected() ) ?"Off" : "On") + "\" /></td>");

    server.print("</tr></table>");
    server.print("</body></html>");
}



void web_resetLog(WebServer &server, WebServer::ConnectionType type, char *, bool)
{
    server.httpSeeOther("/");
    resetLog();
}



void web_reset(WebServer &server, WebServer::ConnectionType type, char *, bool)
{
    server.httpSeeOther("/");
    System.reset();
}



void web_restartModem(WebServer &server, WebServer::ConnectionType type, char *, bool)
{
    server.httpSeeOther("/");
    powerCycleInternet(0);
}



#ifdef SIMULATE_INTERNET_FAILURE
    void web_toggleInternetConnection(WebServer &server, WebServer::ConnectionType type, char *, bool)
    {
        internetFailure ^= 1;
        server.httpSeeOther("/");
    }
#endif



void web_toggleCloudConnection(WebServer &server, WebServer::ConnectionType type, char *, bool)
{
    if ( Particle.connected() ) Particle.disconnect();
    else                        Particle.connect();
    server.httpSeeOther("/");
}



void setup()
{
    byte mac[6];
    char tmp[10];
    struct logEntry logEntry;

    WiFi.on();
    WiFi.connect();

    // Setup the time and store the boot time
    Time.zone(TZ_OFFSET);
    Time.beginDST();
    bootTime = Time.now();
    
    pinMode(LED_PIN,            OUTPUT);
    pinMode(INTERNET_POWER_PIN, OUTPUT);

    digitalWrite(LED_PIN,               HIGH);
    digitalWrite(INTERNET_POWER_PIN,    HIGH);
    
    WiFi.macAddress(mac);
    macAddress="";
    for ( int i = 0; i < 6; i ++ )
    {
        sprintf(tmp, "%02X", mac[i]);
        macAddress += String(tmp);
        if ( i != 5 )   macAddress += ":";
    }
    
    logEntry.entryTime      = Time.now();
    logEntry.entryAction    = LEAT_BOOT;
    addLogEntry(logEntry);
    
    webserver.setDefaultCommand(&web_index);
    webserver.addCommand("resetLog",        &web_resetLog);
    webserver.addCommand("restartModem",    &web_restartModem);
    webserver.addCommand("reset",           &web_reset);

#ifdef SIMULATE_INTERNET_FAILURE
    webserver.addCommand("toggleInternet",    &web_toggleInternetConnection);
#endif

    webserver.addCommand("toggleCloud",    &web_toggleCloudConnection);

    webserver.begin();
}



void loop()
{
    int i, web_len = 64;
    char web_buff[64+1];
    struct logEntry logEntry;
    static int failedPingCount = 0, startupLockout = 1;
    
    if ( ! WiFi.ready() )   return;

    lastPingTime        = Time.now();
    
    // Set startupLockout for the first STARUP_LOCKOUT_TIME seconds since boot, so we don't get a restart at boot
    if ( startupLockout && (lastPingTime >= (bootTime + STARTUP_LOCKOUT_TIME)) )
        startupLockout = 0;
    
    numPacketsReceived  = WiFi.ping(gatewayIP, NUM_PING_PACKETS);

#ifdef SIMULATE_INTERNET_FAILURE
    if ( internetFailure )  numPacketsReceived = 0;
#endif    
    
    // Flash RGB green if we received a successful ping
    if ( numPacketsReceived > 0 )
    {
        if ( failedPingCount )
        {
            logEntry.entryTime      = Time.now();
            logEntry.entryAction    = LEAT_CONNECTION_RESUMED;
            addLogEntry(logEntry);
        }
        
        failedPingCount = 0;
        RGB.control(true);
        RGB.color(255, 255, 255);
        delay(500);
        RGB.control(false);
    }
    else
    {
        if ( failedPingCount < MAX_FAILED_PING_COUNT && ( ! startupLockout) )
        {
            logEntry.entryTime      = Time.now();
            logEntry.entryAction    = LEAT_CONNECTION_LOST;
            addLogEntry(logEntry);
        }

        RGB.control(true);
        RGB.color(255, 0, 0);
        delay(500);
        RGB.control(false);

        failedPingCount ++;
    }
    
    if ( failedPingCount >= MAX_FAILED_PING_COUNT && ( ! startupLockout) )
    {
        powerCycleInternet(1);
    }

    if ( failedPingCount == MAX_FAILED_PING_COUNT && ( ! startupLockout) )
    {
        logEntry.entryTime      = Time.now();
        logEntry.entryAction    = LEAT_DISABLE_CONNECTION_REPORT_TILL_ACTIVE;
        addLogEntry(logEntry);
    }
        
    for ( i = 0; i < 20; i ++ ) // 10seconds
    {
        webserver.processConnection(web_buff, &web_len);
        delay(500);
    }
}
